/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | www.openfoam.com
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Application
    udv

Description
    Tool to extract the velocity profile along a a line which an
    ultrasonic transducer would measure.

Publication
    https://doi.org/10.48550/arXiv.1711.01795
    https://doi.org/10.1016/j.jpowsour.2017.12.042

Developed by
    Vladimir Galindo and Norbert Weber

\*---------------------------------------------------------------------------*/

#include "fvCFD.H"
#include "regionProperties.H"
#include "OFstream.H"

int main(int argc, char *argv[])
{
    #include "options.H" 				// read options
    #include "region.H"
    #include "readData.H"

    List<vector> bp(n);					// points on the beam axis, used for sampling
    List<scalar> distance(n);				// distance of sampling points to sensor
    List<scalar> uMean(n);				// mean velocity at sampling points

    scalar dx = Foam::mag(p2[0] - p[0])/n;		// x-distance between sample points
    scalar dy = Foam::mag(p2[1] - p[1])/n;		// (n does not include the first point)
    scalar dz = Foam::mag(p2[2] - p[2])/n;

    forAll(bp,i)					// compute sample points on beam axis
    {
	bp[i][0] = p[0] + dx + i*dx;			// first point is at dx,dy,dz
        bp[i][1] = p[1] + dy + i*dy;
        bp[i][2] = p[2] + dz + i*dz;
	distance[i] = Foam::mag(bp[i] - p);
    }

    forAll(regionNames, regionI)
    {
        #include "mesh.H"

    	// when searching sampling points, this resolution is used
    	const dimensionedScalar ds("0", dimensionSet(0,1,0,0,0,0,0), Foam::pow(Foam::gMin(mesh.V()),1./3.)/2.0/10.);
    	Info << "    Resolution: " << ds.value()*1000. << " mm" << endl;

	// angle between UDV beam and cell
        volScalarField theta
        (
            IOobject
            (
                "alphaBeam",
                runTime.timeName(),
                mesh,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
	    mesh,
	    dimensionedScalar("0", dimensionSet(0,0,0,0,0,0,0), 0)
        );

	forAll(theta, i)
	{
	    vector dir2 = mesh.C()[i] - p;	// vector from UDV sensor to this point
	    theta[i] = Foam::acos((dir & dir2)/(Foam::mag(dir) * Foam::mag(dir2)));
	}
	theta.write();

        // weighting factor for cells; near field weighted equally in the cylinder
        volScalarField w
        (
            IOobject
            (
                "weightUDV",
                runTime.timeName(),
                mesh,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
            mesh,
            dimensionedScalar("0", dimensionSet(0,0,0,0,0,0,0), 0)
        );

	forAll(w, i)	// calculate sound intensity for each cell
	{
	    dimensionedScalar r("0", dimensionSet(0,1,0,0,0,0,0), Foam::mag(mesh.C()[i]-p));	// distance UDV sensor - this point
	    dimensionedScalar X = pi*D*Foam::sin(theta[i]) / lambda;				// see Krautkraemer p.73
	    w[i] = Foam::sqr(Foam::sin(pi/lambda * (Foam::sqrt(D*D/4.0 + r*r) - r))*Foam::j1(X)/X).value();
	}
	scalar wMax = Foam::gMax(w);			// field will be normalised to a max value of 1.0

        forAll(w, i)					// seach near-field and set all values to 1 inside the beam with diameter of the piezo
        {
	    scalar r = Foam::mag(mesh.C()[i] - p);	// distance to UDV sensor
            double s = r*Foam::sin(theta[i]);		// distance to beam axis

	    if(r < N.value() && s < D.value()/2.0)	// in near field
	    {
		w[i] = 1.0;
	    }
	    else if(r < N.value() && s > D.value()/2.0)	// near, but outside the beam
	    {
                w[i] = 0;
            }
	    else					// normalise to 1 all other cells
	    {
		w[i] /= wMax;
	    }
	}
	w.write();
	Info << nl << "UDV field written" << nl << endl;

    	if(!args.found("s"))
        {
	    Info << "Next entry is the distance of points from the UDV sensor in m:" << nl << "Time = d ";
            forAll(distance,i)
	    {
	    	Info << distance[i] << " ";
	    }
	    Info << endl;
	}

	thetaBeam *= -1; phiBeam *= -1;					// negativ angles for back-transformation

	// select the cells on a sperical plane which belong to each sampling point
    	List<List<label>> cells(n);
    	forAll(cells, i)						// each sampling point
    	{
            double r = distance[i];
            int ns     = (2*r * Foam::sin(theta_cr/2.0) / ds).value();	// number of circles

            for (int is=1; is <= ns; is++)				// each circle (radius=0 is forbidden)
            {
            	double s      = (is*ds).value();			// distance point to beam axis
                double thetaI = 2*Foam::asin(s/(2*r));			// opening angle
		double nphi   = (2.0*pi*s/ds).value();			// number of points on circle

            	for(int iphi=0; iphi <= nphi; iphi++)			// steps on each circle
            	{
                    double phiI   = (2.0*pi*iphi/nphi).value();		// current angle on the circle

		    double x2 = r*Foam::sin(thetaI)*Foam::cos(phiI);
		    double y2 = r*Foam::sin(thetaI)*Foam::sin(phiI);
		    double z2 = r*Foam::cos(thetaI);

		    // move coordinate system from beam axis (above) to real one
         	    double x = p[0] + ((x2*Foam::cos(thetaBeam) -z2*Foam::sin(thetaBeam))*Foam::cos(phiBeam) + y2*Foam::sin(phiBeam)).value();
                    double y = p[1] + (-(x2*Foam::cos(thetaBeam) -z2*Foam::sin(thetaBeam))*Foam::sin(phiBeam) + y2*Foam::cos(phiBeam)).value();
                    double z = p[2] + (x2*Foam::sin(thetaBeam) + z2*Foam::cos(thetaBeam)).value();

//                  double x = p[0] + ((x2*Foam::cos(phiBeam) -y2*Foam::sin(phiBeam))*Foam::cos(thetaBeam) + z2*Foam::sin(thetaBeam)).value();
//                  double y = p[1] + (x2*Foam::sin(phiBeam) + y2*Foam::cos(phiBeam)).value();
//                  double z = p[2] + (-(x2*Foam::cos(phiBeam) - y2*Foam::sin(phiBeam))*Foam::sin(thetaBeam) + z2*Foam::cos(thetaBeam)).value();

		    allPoints << x << ", " << y << ", " << z << endl;
		    if(writeCloud)					// write points to file
                    {
                    	file[i]   << x << ", " << y << ", " << z << endl;
                    }
		    label celli = mesh.findCell(vector(x,y,z));		// find cell with these coordinates
		    if (celli>=0)					// cell found
		    {
			bool found = false;
			forAll(cells[i],ii)
			{
			    if(cells[i][ii] == celli){found = true;} 	// check, if cell is already in the list
			}
			if(found == false){cells[i].append(celli);};
		    }
            	}
            }
    	}

        forAll(timeDirs, timeI)
        {
            runTime.setTime(timeDirs[timeI], timeI);
            mesh.readUpdate();

            IOobject fieldHeader
            (
                "U",
                runTime.timeName(),
                mesh,
                IOobject::MUST_READ 
            );

	    forAll(uMean, i)				// set mean velocity to zero
	    {
		uMean[i]=0;
	    }
          
            // Check if field exists
            if (fieldHeader.typeHeaderOk<volVectorField>(true))
            {
                volVectorField U(fieldHeader, mesh);	// read velocity
		forAll(cells, i)
		{
		    double v = 0;			// volume*weighting factor
		    forAll(cells[i],ii)
		    {
			label c  = cells[i][ii];	// cell id
			vector vec = mesh.C()[c] - p;	
			vector n = vec / (Foam::mag(vec) + SMALL);	// normal vector on spherical surface
			uMean[i] += ((U[c]&n) * mesh.V()[c] * w[c]);	// mean velocity, weighted by cell volume and weighting factor; in beam direction

			v += (mesh.V()[c] * w[c]);
		    }
		    if(v != 0)
		    {
		        uMean[i] /= v;
		    }
		    else
		    {
			uMean[i] = 0;
		    }
		}
		if(args.found("s"))
		{
		    Info << nl << "0 0" << endl;
		    forAll(distance,i)
		    {
		    	Info << distance[i] << " " << uMean[i] << endl;
		    }
		}
		else
		{
		    Info<< "Time = " << runTime.timeName() << " ";
                    forAll(distance,i)
                    {
                        Info << uMean[i] << " ";
                    }
		    Info << endl;
		}
	    }
            else
            {
                Info<< "field not existing" << endl;
            }
    	}
    }
    return 0;
}
